--------------------------------------------------------------------------------
-- Jupiter-80 Librarian
--
-- Copyright 2014 Kim Sand
--
-- SVD utility functions
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Includes
--------------------------------------------------------------------------------
require "utils"
require "mapping"
require "svdcommon"


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------

function part_name(part_type, part_pos)
	local part_type_liveset1 = 0xd4
	local part_type_liveset2 = 0x54
	local part_type_synth1 = 0xdd
	local part_type_synth2 = 0x5d
	local part_type_acoustic1 = 0xd9
	local part_type_acoustic2 = 0x59
	local part_type_acoustic3 = 0xda
	local part_type_acoustic4 = 0x5a
	local part_type_acoustic5 = 0x5c
	local part_type_drumset1 = 0x56
	local part_type_drumset2 = 0xd6

	local part_name = "UNKNOWN"

	if part_type == part_type_liveset1
		or part_type == part_type_liveset2
	then
		local part_pos = (part_pos[1] * 128) + part_pos[2]
		part_name = live_name_table[part_pos + 1]
	elseif part_type == part_type_synth1
		or part_type == part_type_synth2
	then
		local part_pos = (part_pos[1] * 128) + part_pos[2]
		part_name = tone_name_table[part_pos + 1]
	elseif part_type == part_type_acoustic1
		or part_type == part_type_acoustic2
		or part_type == part_type_acoustic3
		or part_type == part_type_acoustic4
		or part_type == part_type_acoustic5
	then
		local bank = string.format("%d %d", part_pos[1], part_pos[2] + 1)
		part_name = reg_tone_mapping[bank]
	elseif part_type == part_type_drumset1
		or part_type == part_type_drumset2
	then
		local bank = string.format("%d %d", part_pos[1], part_pos[2] + 1)
		part_name = reg_drum_mapping[bank]
	end

	return part_name
end

function reg_part_name(part_pos_offset, part_type_offset, text, reg_len, index)
	local part_pos_shifted = {byteidxs(text, part_pos_offset + (reg_len * (index - 1)), 0x02)}
	local part_pos = unshift(part_pos_shifted)

	local part_type = byteidx(text, part_type_offset + (reg_len * (index - 1)))

	local part_name = part_name(part_type, part_pos)

	return part_name, part_pos, part_type
end

function live_part_name(layer_pos_offset, text, live_len, index)
	local layer_meta_shifted = {byteidxs(text, layer_pos_offset + (live_len * (index - 1)), 0x03)}
	local layer_meta = unshift(layer_meta_shifted)

	local layer_pos = {layer_meta[2], layer_meta[3]}
	local layer_type = layer_meta[1]

	local layer_name = part_name(layer_type, layer_pos)

	return layer_name, layer_pos, layer_type
end

function count_text(count, total)
	local count_text

	if count < total then
		count_text = count .. " (" .. total .. ")"
	else
		count_text = total
	end

	return count_text
end

function build_overview()
	-- Check for optional header parts and increase the header offset beyond these

	-- Check for the optional VCL part
	if bytecmp(svd_text, 0x40, {0x56, 0x43, 0x4C}) == true then
		header_offset = header_offset + 0x10
	end

	-- Check for the optional SYS part
	if bytecmp(svd_text, 0x50, {0x53, 0x59, 0x53}) == true then
		header_offset = header_offset + 0x10
	end

	-- Check for the optional RBN part
	if bytecmp(svd_text, 0x60, {0x52, 0x42, 0x4E}) == true then
		header_offset = header_offset + 0x10
	end

	reg_offset = reg_offset + header_offset


	-- Get the counts for number of registrations, live sets and tones

	nr_of_regs = number_from_bytes(svd_text, nr_of_regs_offset + header_offset, 4)
	nr_of_lives_offset = reg_offset + (nr_of_regs * reg_len)
	nr_of_lives = number_from_bytes(svd_text, nr_of_lives_offset, 4)
	live_offset = nr_of_lives_offset + 4 + live_meta_len
	nr_of_tones_offset = live_offset + (nr_of_lives * live_len)
	nr_of_tones = number_from_bytes(svd_text, nr_of_tones_offset, 4)
	tone_offset = nr_of_tones_offset + 4 + tone_meta_len

	renoise.app():show_status("Building overview for " .. nr_of_regs .. " registrations, " .. nr_of_lives .. " live sets and " .. nr_of_tones .. " tones.")


	-- Get the names of all registrations, live sets and tones

	local reg_count = 0

	local reg_names = "Registration names:"

	for index=1, nr_of_regs do
		local reg_name_shifted = {byteidxs(svd_text, reg_offset + (reg_len * (index - 1)), 0x10)}
		local reg_unshifted = unshift(reg_name_shifted)
		local reg_name = string_from_table(reg_unshifted)
		reg_name = trimr(reg_name)

		-- Ignore uninitialized registrations
		if reg_name ~= init_reg_name then
			reg_names = reg_names .. "\n" .. reg_name
			reg_count = reg_count + 1
		end

		table.insert(reg_name_table, reg_name)
	end

	local live_count = 0

	local live_names = "Live set names:"

	for index=1, nr_of_lives do
		local live_name_shifted = {byteidxs(svd_text, live_offset + (live_len * (index - 1)), 0x10)}
		local live_unshifted = unshift(live_name_shifted)
		local live_name = string_from_table(live_unshifted)
		live_name = trimr(live_name)

		-- Ignore uninitialized live sets
		if live_name ~= init_live_name then
			live_names = live_names .. "\n" .. live_name
			live_count = live_count + 1
		end

		table.insert(live_name_table, live_name)
	end

	local tone_count = 0

	local tone_names = "Tone names:"

	for index=1, nr_of_tones do
		local tone_name_shifted = {byteidxs(svd_text, tone_offset + (tone_len * (index - 1)), 0x0C)}
		local tone_unshifted = unshift(tone_name_shifted)
		local tone_name = string_from_table(tone_unshifted)
		tone_name = trimr(tone_name)

		-- Ignore uninitialized tones
		if tone_name ~= init_tone_name then
			tone_names = tone_names .. "\n" .. tone_name
			tone_count = tone_count + 1
		end

		table.insert(tone_name_table, tone_name)
	end


	-- Build an overview text with the names and counts

	local reg_count_text = count_text(reg_count, nr_of_regs)
	local live_count_text = count_text(live_count, nr_of_lives)
	local tone_count_text = count_text(tone_count, nr_of_tones)

	local count_text = "Nr of registrations:" .. reg_count_text .. "\n"
	count_text = count_text .. "Nr of live sets:" .. live_count_text .. "\n"
	count_text = count_text .. "Nr of tones:" .. tone_count_text

	local name_text = reg_names .. "\n\n" .. live_names .. "\n\n" .. tone_names

	overview_text = count_text .. "\n\n" .. name_text
end

function build_independents()
	local live_names = "Unused live set names:\n"

	local nr_of_lives = 0
	local live_count = 0

	for index=1, #live_name_table do
		if live_reg_map[index] == nil then
			local live_name = live_name_table[index]

			-- Ignore uninitialized live sets
			if live_name ~= init_live_name then
				live_names = live_names .. live_name .. "\n"
				live_count = live_count + 1
			end

			nr_of_lives = nr_of_lives + 1
		end
	end

	if live_count <= 0 then
		live_names = live_names .. "None\n"
	end

	local tone_names = "Unused tones:\n"

	local nr_of_tones = 0
	local tone_count = 0

	for index=1, #tone_name_table do
		if tone_reg_map[index] == nil
			and tone_live_map[index] == nil
		then
			local tone_name = tone_name_table[index]

			-- Ignore uninitialized tones
			if tone_name ~= init_tone_name then
				tone_names = tone_names .. tone_name .. "\n"
				tone_count = tone_count + 1
			end

			nr_of_tones = nr_of_tones + 1
		end
	end

	if tone_count <= 0 then
		tone_names = tone_names .. "None\n"
	end


	-- Build a text with the names and counts

	local live_count_text = count_text(live_count, nr_of_lives)
	local tone_count_text = count_text(tone_count, nr_of_tones)

	local count_text = "Nr of unused live sets:" .. live_count_text .. "\n"
	count_text = count_text .. "Nr of unused tones:" .. tone_count_text

	local name_text = live_names .. "\n" .. tone_names

	independents_text = count_text .. "\n\n" .. name_text
end

function build_dependencies()
	renoise.app():show_status("Building dependencies for " .. nr_of_regs .. " registrations, " .. nr_of_lives .. " live sets and " .. nr_of_tones .. " tones.")

	-- Build a text with live set dependencies

	local live_layer1_offset = 0x19f
	local live_layer2_offset = 0x1c5
	local live_layer3_offset = 0x1eb
	local live_layer4_offset = 0x211

	local live_dep_text = "Live set name,Live set layer 1,Live set layer 2,Live set layer 3,Live set layer 4\n"

	local live_dep_table = {}

	for index=1, nr_of_lives do
		live_dep_text = live_dep_text .. live_name_table[index]

		local layer_dep_text = ""

		local layer1_name, layer1_pos = live_part_name(live_offset + live_layer1_offset, svd_text, live_len, index)
		layer_dep_text = layer_dep_text .. "," .. layer1_name

		local layer2_name, layer2_pos = live_part_name(live_offset + live_layer2_offset, svd_text, live_len, index)
		layer_dep_text = layer_dep_text .. "," .. layer2_name

		local layer3_name, layer3_pos = live_part_name(live_offset + live_layer3_offset, svd_text, live_len, index)
		layer_dep_text = layer_dep_text .. "," .. layer3_name

		local layer4_name, layer4_pos = live_part_name(live_offset + live_layer4_offset, svd_text, live_len, index)
		layer_dep_text = layer_dep_text .. "," .. layer4_name

		table.insert(live_dep_table, layer_dep_text)

		live_dep_text = live_dep_text .. layer_dep_text .. "\n"


		-- Build an inverse map from tones to the registrations that use them

		layer1_pos = (layer1_pos[1] * 128) + layer1_pos[2]
		layer2_pos = (layer2_pos[1] * 128) + layer2_pos[2]
		layer3_pos = (layer3_pos[1] * 128) + layer3_pos[2]
		layer4_pos = (layer4_pos[1] * 128) + layer4_pos[2]

		local live_tone_table = {layer1_pos, layer2_pos, layer3_pos, layer4_pos}
		local live_offset_table = {live_layer1_offset, live_layer2_offset, live_layer3_offset, live_layer4_offset}

		create_inverse_map(live_tone_table, live_offset_table, tone_live_map, index)
	end


	-- Build a text with registration dependencies

	reg_upper_offset = reg_upper_offset + header_offset
	reg_lower_offset = reg_lower_offset + header_offset
	reg_solo_offset = reg_solo_offset + header_offset
	reg_perc_offset = reg_perc_offset + header_offset

	reg_upper_type_offset = reg_upper_type_offset + header_offset
	reg_lower_type_offset = reg_lower_type_offset + header_offset
	reg_solo_type_offset = reg_solo_type_offset + header_offset
	reg_perc_type_offset = reg_perc_type_offset + header_offset

	local reg_dep_text

	if file_format == File_format.Jupiter80 then
		reg_dep_text = "Registration name,Upper live set name,Lower live set name, Solo part name, Perc part name,Upper live set layer 1,2,3,4,Lower live set layer 1,2,3,4\n"
	elseif file_format == File_format.Jupiter50 then
		reg_dep_text = "Registration name,Live set name,Solo part name, Perc part name,Live set layer 1,2,3,4\n"
	end

	for index=1, nr_of_regs do
		reg_dep_text = reg_dep_text .. reg_name_table[index]

		local upper_name, reg_upper_pos = reg_part_name(reg_upper_offset, reg_upper_type_offset, svd_text, reg_len, index)
		reg_dep_text = reg_dep_text .. "," .. upper_name

		local reg_lower_pos, lower_name

		if file_format == File_format.Jupiter80 then
			lower_name, reg_lower_pos = reg_part_name(reg_lower_offset, reg_lower_type_offset, svd_text, reg_len, index)
			reg_dep_text = reg_dep_text .. "," .. lower_name
		end

		local solo_name, reg_solo_pos = reg_part_name(reg_solo_offset, reg_solo_type_offset, svd_text, reg_len, index)
		reg_dep_text = reg_dep_text .. "," .. solo_name

		local perc_name, reg_perc_pos = reg_part_name(reg_perc_offset, reg_perc_type_offset, svd_text, reg_len, index)
		reg_dep_text = reg_dep_text .. "," .. perc_name

		reg_upper_pos = (reg_upper_pos[1] * 128) + reg_upper_pos[2]

		local upper_layer_text = live_dep_table[reg_upper_pos + 1]
		reg_dep_text = reg_dep_text .. upper_layer_text

		if file_format == File_format.Jupiter80 then
			reg_lower_pos = (reg_lower_pos[1] * 128) + reg_lower_pos[2]

			local lower_layer_text = live_dep_table[reg_lower_pos + 1]
			reg_dep_text = reg_dep_text .. lower_layer_text
		end

		reg_dep_text = reg_dep_text .. "\n"


		-- Build an inverse map from live sets to the registrations that use them

		local reg_live_table = {reg_upper_pos, reg_lower_pos}
		local reg_offset_table = {reg_upper_type_offset, reg_lower_type_offset}

		create_inverse_map(reg_live_table, reg_offset_table, live_reg_map, index)

		-- Build an inverse map from tones to the registrations that use them

		reg_solo_pos = (reg_solo_pos[1] * 128) + reg_solo_pos[2]
		reg_perc_pos = (reg_perc_pos[1] * 128) + reg_perc_pos[2]

		local reg_tone_table = {reg_solo_pos, reg_perc_pos}
		local reg_offset_table = {reg_solo_type_offset, reg_perc_type_offset}

		create_inverse_map(reg_tone_table, reg_offset_table, tone_reg_map, index)
	end


	-- Build a dependencies text with registrations and live sets

	dependencies_text = reg_dep_text .. "\n" .. live_dep_text
end

function build_inverse_dependencies()
	inv_dependencies_text = ""

	local live_reg_text = "Live set name, registration names\n"
	build_inverse_dep(live_reg_map, live_name_table, reg_name_table, live_reg_text)

	local tone_reg_text = "\nTone name, registration names\n"
	build_inverse_dep(tone_reg_map, tone_name_table, reg_name_table, tone_reg_text)

	local tone_live_text = "\nTone name, live set names\n"
	build_inverse_dep(tone_live_map, tone_name_table, live_name_table, tone_live_text)
end

function build_inverse_dep(dep_map, from_name_table, to_name_table, dep_header)
	local part_table = {}

	for key,value in pairs(dep_map) do
		table.insert(part_table, {key, value})
	end

	table.sort(part_table, inv_map_compare)

	inv_dependencies_text = inv_dependencies_text .. dep_header

	for index=1, #part_table do
		local map_entry = part_table[index]

		local from_nr = map_entry[1]
		local from_name = from_name_table[from_nr]

		-- The name is not found when referring to acoustic tones
		-- Ignore uninitialized sounds
		if from_name ~= nil
			and from_name ~= init_live_name
			and from_name ~= init_tone_name
		then
			local to_name_text = ""
			local to_nrs = map_entry[2]

			for index=1, #to_nrs do
				local to_nr = to_nrs[index][1]
				local to_name = to_name_table[to_nr]

				-- Ignore uninitialized sounds
				if to_name ~= init_reg_name
					and to_name ~= init_live_name
				then
					to_name_text = to_name_text .. "," .. to_name
				end
			end

			if #to_name_text > 0 then
				inv_dependencies_text = inv_dependencies_text .. from_name .. to_name_text .. "\n"
			end
		end
	end
end

function inv_map_compare(map_a, map_b)
	return #map_a[2] > #map_b[2]
end

function create_inverse_map(part_table, offset_table, map_table, index)
	-- map_table format example (live-to-reg map):
	-- {live_index, {{reg_index, reg_part_offset}, ...}}
	for part_i=1, #part_table do
		local part_pos = part_table[part_i] + 1
		local part_offset = offset_table[part_i]

		if map_table[part_pos] == nil then
			map_table[part_pos] = {{index, part_offset}}
		else
			table.insert(map_table[part_pos], {index, part_offset})
		end
	end
end

function swap_registrations()
	if swap_from_part ~= nil and swap_to_part ~= nil then
		-- Swap positions in the structure
		local pos1 = reg_offset + (reg_len * (swap_from_part - 1))
		local pos2 = reg_offset + (reg_len * (swap_to_part - 1))

		svd_text = swap_chars(svd_text, pos1, reg_len, pos2, reg_len)

		swap_from_part = nil
		swap_to_part = nil
	end
end

function pos_from_inverse_map(inv_map, index, part_len)
	local part_index = inv_map[index][1]
	local part_offset = inv_map[index][2]

	local pos = part_offset + (part_len * (part_index - 1))

	return pos
end

function replace_part_bytes_in_inverse_map(inv_map, index, part_len, part_pos)
	local pos = pos_from_inverse_map(inv_map, index, part_len)

	local part_bytes_shifted = {byteidxs(svd_text, pos+1, 0x06)}
	--print("part_bytes_shifted")
	--print_table_hex(part_bytes_shifted)

	local part_bytes_unshifted = unshift(part_bytes_shifted)
	--print("part_bytes_unshifted")
	--print_table_hex(part_bytes_unshifted)

	local bytes = bytes_from_number(part_pos)
	--print("bytes")
	--print_table_hex(bytes)

	if #bytes > 1 then
		part_bytes_unshifted[1] = bytes[1]
		part_bytes_unshifted[2] = bytes[2]
	else
		part_bytes_unshifted[1] = 0
		part_bytes_unshifted[2] = bytes[1]
	end

	local part_bytes_reshifted = shift(part_bytes_unshifted)

	-- The sound type byte is not shifted, and can be prepended unmodified
	local type_byte = byteidx(svd_text, pos)
	table.insert(part_bytes_reshifted, 1, type_byte)

	--print("part_bytes_reshifted")
	--print_table_hex(part_bytes_reshifted)

	return part_bytes_reshifted
end

function swap_live_sets()
	if swap_from_part ~= nil and swap_to_part ~= nil then
		-- Swap positions in the structure

		local pos1 = live_offset + (live_len * (swap_from_part - 1))
		local pos2 = live_offset + (live_len * (swap_to_part - 1))

		svd_text = swap_chars(svd_text, pos1, live_len, pos2, live_len)

		-- Update positions in registrations: Unshift, update and shift back
		local from_regs = live_reg_map[swap_from_part]

		if from_regs ~= nil then
			for reg_i=1, #from_regs do
				local part_bytes = replace_part_bytes_in_inverse_map(from_regs, reg_i, reg_len, swap_to_part - 1)
				local from_pos = pos_from_inverse_map(from_regs, reg_i, reg_len)
				svd_text = replace_chars(svd_text, from_pos, part_bytes)
			end
		end

		local to_regs = live_reg_map[swap_to_part]

		if to_regs ~= nil then
			for reg_i=1, #to_regs do
				local part_bytes = replace_part_bytes_in_inverse_map(to_regs, reg_i, reg_len, swap_from_part - 1)
				local to_pos = pos_from_inverse_map(to_regs, reg_i, reg_len)
				svd_text = replace_chars(svd_text, to_pos, part_bytes)
			end
		end

		-- Swap names in the live set name table
		live_name_table[swap_from_part], live_name_table[swap_to_part] = live_name_table[swap_to_part], live_name_table[swap_from_part]

		swap_from_part = nil
		swap_to_part = nil
	end
end

function swap_tones()
	if swap_from_part ~= nil and swap_to_part ~= nil then
		-- Swap positions in the structure

		local pos1 = tone_offset + (tone_len * (swap_from_part - 1))
		local pos2 = tone_offset + (tone_len * (swap_to_part - 1))

		svd_text = swap_chars(svd_text, pos1, tone_len, pos2, tone_len)

		-- Update positions in registrations: Unshift, update and shift back
		local from_regs = tone_reg_map[swap_from_part]

		if from_regs ~= nil then
			for reg_i=1, #from_regs do
				local part_bytes = replace_part_bytes_in_inverse_map(from_regs, reg_i, reg_len, swap_to_part - 1)
				local from_pos = pos_from_inverse_map(from_regs, reg_i, reg_len)
				svd_text = replace_chars(svd_text, from_pos, part_bytes)
			end
		end

		local to_regs = tone_reg_map[swap_to_part]

		if to_regs ~= nil then
			for reg_i=1, #to_regs do
				local part_bytes = replace_part_bytes_in_inverse_map(to_regs, reg_i, reg_len, swap_from_part - 1)
				local to_pos = pos_from_inverse_map(to_regs, reg_i, reg_len)
				svd_text = replace_chars(svd_text, to_pos, part_bytes)
			end
		end

		-- Update positions in live sets: Unshift, update and shift back
		local from_lives = tone_live_map[swap_from_part]

		-- TODO: The live set meta offsets and sizes are different from the registration ones

		if from_lives ~= nil then
			for live_i=1, #from_lives do
				local part_bytes = replace_part_bytes_in_inverse_map(from_lives, live_i, live_len, swap_to_part - 1)
				local from_pos = pos_from_inverse_map(from_lives, live_i, live_len)
				svd_text = replace_chars(svd_text, from_pos, part_bytes)
			end
		end

		local to_lives = tone_live_map[swap_to_part]

		if to_lives ~= nil then
			for live_i=1, #to_lives do
				local part_bytes = replace_part_bytes_in_inverse_map(to_lives, live_i, live_len, swap_from_part - 1)
				local to_pos = pos_from_inverse_map(to_lives, live_i, live_len)
				svd_text = replace_chars(svd_text, to_pos, part_bytes)
			end
		end

		-- Swap names in the tone name table
		tone_name_table[swap_from_part], tone_name_table[swap_to_part] = tone_name_table[swap_to_part], tone_name_table[swap_from_part]

		swap_from_part = nil
		swap_to_part = nil
	end
end