--------------------------------------------------------------------------------
-- Jupiter-80 Librarian
--
-- Copyright 2014 Kim Sand
--
-- Sysex and MIDI communication
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Includes
--------------------------------------------------------------------------------
require "mapping"
require "utils"
require "midi"


--------------------------------------------------------------------------------
-- Variables
--------------------------------------------------------------------------------

Loop_type = {Registrations=0, Livesets=1}

sysex_buffer = nil
message_start = nil
message_end = nil
message_type = nil
message_buffer = nil
dependencies_text = nil
registration_pc = nil
registration_pc_max = nil
liveset_pc = nil
liveset_pc_max = nil
loop_type = nil


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------

function init_variables()
  sysex_buffer = {}
  message_start = 1
  message_end = 19
  message_type = message_start
  message_buffer = ""
  registration_pc = 0
  registration_pc_max = 255
  liveset_pc = 0
  liveset_pc_max = 2559

  if loop_type == Loop_type.Livesets then
    dependencies_text = "Live set name,Live set layer 1,2,3,4\n"
  else
    dependencies_text = "Registration name,Upper live set name,Lower live set name,Upper live set layer 1,2,3,4,Lower live set layer 1,2,3,4\n"
  end
end

function connect_to_jupiter(loop_type)
  init_variables()

  local found = false
  local wrong_mode = false

  local out_devices = renoise.Midi.available_output_devices()
  local in_devices = renoise.Midi.available_input_devices()

  local jupiter_device = "Roland JUPITER-80"
  local vender_device = "JUPITER-80"

  for i = 1,#in_devices do
    if in_devices[i] == jupiter_device then
      found = true
    end

    if in_devices[i] == vender_device then
      wrong_mode = true
    end
  end

  if found == true then
    -- Found input device, connect!
    midi_in_connect(jupiter_device)
  else
    if wrong_mode == true then
      renoise.app():show_error("Vender MIDI+AUDIO mode not supported. Set Jupiter-80 to generic MIDI mode.")
    else
      renoise.app():show_error("Could not locate " .. jupiter_device .. " as MIDI input device.")
    end

    return false
  end

  for i = 1,#out_devices do
    if out_devices[i] == jupiter_device then
      found = true
    end

    if out_devices[i] == vender_device then
      wrong_mode = true
    end
  end

  if found == true then
    -- Found output device, connect!
    midi_out_connect(jupiter_device)
  else
    if wrong_mode == true then
      renoise.app():show_error("Vender MIDI+AUDIO mode not supported. Set Jupiter-80 to generic MIDI mode.")
    else
      renoise.app():show_error("Could not locate " .. jupiter_device .. " as MIDI output device.")
    end

    return false
  end

  return true
end

function send_jupiter_commands()
  -- Run through all commands in order and loop until finished.
  -- One iteration fetches all data for one registration or live set.
  -- Each iteration starts by changing to the next registration or live set.
  if message_type == 1 then
    -- If looping through live sets
    if loop_type == Loop_type.Livesets then
      -- Build a command to change to the next live set
      local bank_lsb = math.floor(liveset_pc / 128)
      local program_nr = liveset_pc % 128
      set_upper_liveset_pc[12] = bank_lsb
      set_upper_liveset_pc[13] = program_nr

      local remainder = (122 + bank_lsb + program_nr) % 128
      local checksum = 128 - remainder
      set_upper_liveset_pc[16] = checksum

      --print("set_upper_liveset_pc:", unpack(set_upper_liveset_pc))

      -- Perform the program change
      midi_out:send(set_upper_liveset_pc)

      message_type = 2
      midi_out:send(req_upper_liveset)
    -- Else when looping through registrations
    else
      -- Build a command to change to the next registration
      local bank_lsb = math.floor(registration_pc / 128)
      local program_nr = registration_pc % 128
      set_registration_pc[12] = bank_lsb
      set_registration_pc[13] = program_nr

      local remainder = (2 + bank_lsb + program_nr) % 128
      local checksum = 128 - remainder
      set_registration_pc[16] = checksum

      --print("set_registration_pc:", unpack(set_registration_pc))

      -- Perform the program change
      midi_out:send(set_registration_pc)

      midi_out:send(req_registration)
    end
  elseif message_type == 2 then
    midi_out:send(req_upper_liveset)
  elseif message_type == 3 then
    midi_out:send(req_lower_liveset)
  elseif message_type == 4 then
    midi_out:send(req_upper_layer1_bank)
  elseif message_type == 5 then
    midi_out:send(req_upper_layer1)
  elseif message_type == 6 then
    midi_out:send(req_upper_layer2_bank)
  elseif message_type == 7 then
    midi_out:send(req_upper_layer2)
  elseif message_type == 8 then
    midi_out:send(req_upper_layer3_bank)
  elseif message_type == 9 then
    midi_out:send(req_upper_layer3)
  elseif message_type == 10 then
    midi_out:send(req_upper_layer4_bank)
  elseif message_type == 11 then
    midi_out:send(req_upper_layer4)
  elseif message_type == 12 then
    midi_out:send(req_lower_layer1_bank)
  elseif message_type == 13 then
    midi_out:send(req_lower_layer1)
  elseif message_type == 14 then
    midi_out:send(req_lower_layer2_bank)
  elseif message_type == 15 then
    midi_out:send(req_lower_layer2)
  elseif message_type == 16 then
    midi_out:send(req_lower_layer3_bank)
  elseif message_type == 17 then
    midi_out:send(req_lower_layer3)
  elseif message_type == 18 then
    midi_out:send(req_lower_layer4_bank)
  elseif message_type == 19 then
    midi_out:send(req_lower_layer4)
  end
end

function my_sysex_callback(message)
  --print("message_type:", message_type)

  table.insert(sysex_buffer, message)

  if loop_type == Loop_type.Livesets then
    renoise.app():show_status("Retrieving " .. liveset_pc + 1 .. " of " .. liveset_pc_max + 1 .. " live sets.")
  else
    renoise.app():show_status("Retrieving " .. registration_pc + 1 .. " of " .. registration_pc_max + 1 .. " registrations.")
  end

  --renoise.app():show_status("SysEx Handler received " .. #message .. " bytes.")
  --print(("Got SYSEX with %d bytes"):format(#message))

  local last_byte = message[#message]

  -- When the last byte is 0xF7 the message is complete
  if last_byte == 0xF7 then
    -- Flatten buffer to a one-dimension table
    local flat_buffer = {}
    for i = 1,#sysex_buffer do
      for j = 1,#sysex_buffer[i] do
         table.insert(flat_buffer, sysex_buffer[i][j])
      end
    end

    sysex_buffer = {}

    if #flat_buffer > 1 then
      local name = ""
      local tone_name = nil

      -- If the message is a bank
      if message_type == 4
        or message_type == 6
        or message_type == 8
        or message_type == 10
        or message_type == 12
        or message_type == 14
        or message_type == 16
        or message_type == 18
        then

        -- Check if the bank corresponds with an acoustic (non-synth) tone
        local bank = string.format("%d %d %d", message[12], message[13], message[14] + 1)
        tone_name = tone_mapping[bank]

        -- Use a predefined name for acoustic (non-synth) tones
        if tone_name ~= nil then
          name = tone_name
          message_type = message_type + 1
        end
      else
        local end_index

        -- Tones have shorter names than registrations and live sets (12 chars vs 16)
        if message_type < 4 then
          end_index = 27
        else
          end_index = 23
        end

        -- Build the name from the list of characters
        for i = 12, end_index do
           name = name .. string.char(message[i])
        end
      end

      -- Trim whitespace from the right side of the name
      if name ~= "" then
        if message_buffer == "" then
          message_buffer = trimr(name)
        else
          message_buffer = message_buffer .. "," .. trimr(name)
        end
      end

      -- When looping through live sets
      if loop_type == Loop_type.Livesets then
        -- Skip the lower live set
        if message_type == 2 then
          message_type = message_type + 1
        -- Skip the lower live set layers
        elseif message_type == 11 then
          message_type = message_end
        end
      end

      -- Continue through all commands in order
      if message_type < message_end then
        message_type = message_type + 1
        send_jupiter_commands()
      -- When all commands are sent
      else
        -- Print all data for the current registration or live set
        --print(message_buffer)

        -- Build up a text wih one line for each registration or live set
        dependencies_text = dependencies_text .. message_buffer .. "\n"

        local do_continue_loop = false

        -- Check if there are more registrations or live sets to loop through
        if loop_type == Loop_type.Livesets then
          if liveset_pc < liveset_pc_max then
            liveset_pc = liveset_pc + 1
            do_continue_loop = true
          end
        else
          if registration_pc < registration_pc_max then
            registration_pc = registration_pc + 1
            do_continue_loop = true
          end
        end

        -- If continuing the loop, wrap around to the first command
        if do_continue_loop == true then
          message_type = message_start
          message_buffer = ""
          send_jupiter_commands()
        -- Otherwise disconnect from the synthesizer
        else
          midi_in_disconnect()
          midi_out_disconnect()

          if loop_type == Loop_type.Livesets then
            renoise.app():show_status("Done retrieving live sets.")
          else
            renoise.app():show_status("Done retrieving registrations.")
          end

          show_result()
        end
      end
    end
  end
end

function prompt_save_list_from_sysex()
  local filename = renoise.app():prompt_for_filename_to_write(".csv",
    "Save list as")

  if filename == "" then
    renoise.app():show_status("File saving cancelled.")
    return false
  else
    local f_out = io.open(filename, "wb")

    if f_out == nil then
      renoise.app():show_status("Couldn't open file for writing: " .. filename .. ".")
      return
    end

    f_out:seek("set")
    f_out:write(dependencies_text)
    f_out:flush()
    f_out:close()

    renoise.app():show_status("File saved successfully as: " .. filename .. ".")
  end
end

function show_result()
  local vb = renoise.ViewBuilder()

  local DIALOG_MARGIN = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
  local DIALOG_SPACING = renoise.ViewBuilder.DEFAULT_DIALOG_SPACING
  local CONTROL_MARGIN = renoise.ViewBuilder.DEFAULT_CONTROL_MARGIN

  local multiline_text = vb:column {
    margin = CONTROL_MARGIN,
    vb:button {
      text = "Save list to file ...",
      notifier = prompt_save_list_from_sysex
    },
    vb:multiline_text{
      width = 800,
      height = 800,
      text = dependencies_text
    }
  }

  local dialog_content = vb:column {
      margin = DIALOG_MARGIN,
      spacing = DIALOG_SPACING,
      uniform = true,

      multiline_text
  }

  local dialog_title

  if loop_type == Loop_type.Livesets then
    dialog_title = "Retrieved live set names"
  else
    dialog_title = "Retrieved registration names"
  end

  local dialog = renoise.app():show_custom_dialog(
    dialog_title,
    dialog_content
  )

end
