--------------------------------------------------------------------------------
-- Jupiter-80 Librarian
--
-- Copyright 2014 Kim Sand
--
-- Utility functions
--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Functions
--------------------------------------------------------------------------------

package.path = 'lmod/?.lua;' .. package.path
local bit = require 'bit.numberlua'

function pack(...)
	return arg
end

function table_slice (values,i1,i2)
	local res = {}
	local n = #values
	-- default values for range
	i1 = i1 or 1
	i2 = i2 or n
	if i2 < 0 then
		i2 = n + i2 + 1
	elseif i2 > n then
		i2 = n
	end
	if i1 < 1 or i1 > n then
		return {}
	end
	local k = 1
	for i = i1,i2 do
		res[k] = values[i]
		k = k + 1
	end
	return res
end

function splitfilenamepath(strfilename)
	-- Return the filename only from the path
	local match = string.match(strfilename, "[\\/]([^/\\]+)$")
	return match
end

function trimr(s)
	-- Trim whitespace from the right end of the string
	return s:find'^%s*$' and '' or s:match'^(.*%S)'
end

function byteidx(text, index)
	return text:sub(index + 1):byte()
end

function byteidxs(text, index, count)
	return text:sub(index + 1, index + count):byte(1,count)
end

function bytecmp(text, start, bytelist)
	local is_equal = true

	local bytes = {byteidxs(text, start, #bytelist)}
	for index=1, #bytes do
		if bytes[index] ~= bytelist[index] then
			is_equal = false
			break
		end
	end

	return is_equal
end

function number_from_bytes(text, start, count)
	local number = byteidx(text, start + count - 1)

	local iteration = count - 2

	for index = start, start + count - 2 do
		local byte = byteidx(text, index)
		number = number + byte * (2^(8+iteration))

		iteration = iteration - 1
	end

	return number
end

function bytes_from_number(number)
	local count = math.ceil(select(2,math.frexp(number))/8)

	local bytes = {}

	if number == 0 then
		bytes[1] = 0
	else
		for index = 1, count do
			local multiplier = 2^(8*(index-1))
			bytes[index] = math.floor(number/multiplier)
			number = number - bytes[index] * multiplier
		end
	end

	return bytes
end

function string_from_table(source_table)
	local dest_string = ""

	for index=1, #source_table do
		dest_string = dest_string .. string.char(source_table[index])
	end

	return dest_string
end

function table_join(table1, table2)
	for index=1, #table2 do
		table.insert(table1, table2[index])
	end

	return table1
end

function print_table(a_table)
	for key,value in pairs(a_table) do print(key,value) end
end

function print_table_hex(a_table)
	for key,value in pairs(a_table) do
		value = string.format("%02x", value)
		print(key,value)
	end
end

function swap_chars(text, pos1, len1, pos2, len2)
	-- Ensure the earliest position in the string is handled first
	if pos1 > pos2 then
		pos1, pos2 = pos2, pos1
		len1, len2 = len2, len1
	end

	local pre1 = text:sub(1, pos1-1)
	local text1 = text:sub(pos1, pos1+len1-1)
	local pre2 = text:sub(pos1+len1, pos2-1)
	local text2 = text:sub(pos2, pos2+len2-1)
	local post = text:sub(pos2+len2)

	return table.concat{pre1, text2, pre2, text1, post}
end

function replace_chars(text, pos, chars)
	local pre = text:sub(1, pos)
	local post = text:sub(pos+1+#chars)
	local char_str = string_from_table(chars)

	return table.concat{pre, char_str, post}
end

-- 7 bytes shifted contain 8 bytes of information unshifted.
-- We only need to read 7 bytes (shifted) to produce 8 (unshifted).
-- Instead of unshifting the bytes, the 7 bits can be extracted by offset.
-- We need to read at most two bytes each step to extract all 7 bits.
-- Each step we increase the byte offset by 1 (reading two bytes).
-- Each step we decrease the bit offset by 1.
-- In eight steps the bit offset has wrapped around.
--
-- These are the steps needed to extract 8 bytes of information:
-- 11111110 00000000 00000000 00000000 00000000 00000000 00000000 00000000
-- 00000001 11111100 00000000 00000000 00000000 00000000 00000000 00000000
-- 00000000 00000011 11111000 00000000 00000000 00000000 00000000 00000000
-- 00000000 00000000 00000111 11110000 00000000 00000000 00000000 00000000
-- 00000000 00000000 00000000 00001111 11100000 00000000 00000000 00000000
-- 00000000 00000000 00000000 00000000 00011111 11000000 00000000 00000000
-- 00000000 00000000 00000000 00000000 00000000 00111111 10000000 00000000
-- 00000000 00000000 00000000 00000000 00000000 00000000 01111111 00000000

function shift(bytes)
	-- 7 bytes shifted contain 8 bytes of information unshifted.
	-- For each eight byte, decrement the bytes needed by 1.
	local bytes_decrement = math.floor(#bytes / 8)
	local bytes_needed = #bytes - bytes_decrement

	local shifted = {}

	-- Create an empty table to be filled out with shifted bytes
	for index=1, bytes_needed do
		table.insert(shifted, 0)
	end

	local bit_offset = 0

	for index=1, #bytes do
		local firstbyte, secondbyte = extract_two_bytes(index, shifted)
		local hexstr = hexstr_from_two_bytes(firstbyte, secondbyte)
		--print("hexstr before:", hexstr)

		local byte = bit.lshift(bytes[index], 1)
		local bytebits = bit.replace(hexstr, byte, bit_offset, 8)
		local firstbits = bit.extract(bytebits, 8, 8)
		local secondbits = bit.extract(bytebits, 0, 8)
		--print("hexstr after:", string.format("0x%02x%02x", firstbits, secondbits))

		if index == 1 then
			shifted[index] = secondbits
		else
			shifted[index - 1] = firstbits
			shifted[index] = secondbits
		end

		-- On each eight byte the bit offset wraps around
		if bit_offset < 7 then
			bit_offset = bit_offset + 1
		else
			bit_offset = 0
		end
	end

	return shifted
end

function unshift(bytes)
	local unshifted = {}

	local bit_offset = 0

	for index=1, #bytes do
		-- Extract 2 bytes, which is all we need for extracting 7 bits.
		-- The bit functions are limited to maximum 4 bytes, so the split is required.
		local firstbyte, secondbyte = extract_two_bytes(index, bytes)
		local hexstr = hexstr_from_two_bytes(firstbyte, secondbyte)

		-- Extract 8 bits to the left of the bit offset
		local bytebits = bit.extract(hexstr, bit_offset, 8)

		-- Shift the bits 1 position to the right to get the 7 bit value
		local byte = bit.rshift(bytebits, 1)

		-- Add the unshifted byte to the unshifted list
		table.insert(unshifted, byte)

		-- On each eight byte the bit offset wraps around
		if bit_offset < 7 then
			bit_offset = bit_offset + 1
		else
			bit_offset = 0
		end
	end

	return unshifted
end

function extract_two_bytes(index, bytes)
	local firstbyte
	local secondbyte

	-- 7 bytes shifted contain 8 bytes of information unshifted.
	-- For each eight byte, decrement the index by 1.
	local index_decrement = math.floor((index - 1) / 8)
	index = index - index_decrement

	-- The first byte has no previous byte and must be padded with a zero-byte
	if index == 1 then
		firstbyte = 0
		secondbyte = bytes[index]
	else
		firstbyte = bytes[index - 1]
		secondbyte = bytes[index]
	end

	return firstbyte, secondbyte
end

function hexstr_from_two_bytes(firstbyte, secondbyte)
	-- Create a hexadecimal string representation of the two bytes
	local hexstr = "0x"
	hexstr = hexstr .. string.format("%02x", firstbyte)
	hexstr = hexstr .. string.format("%02x", secondbyte)

	return hexstr
end
